# -*- coding: utf-8 -*-

XML_NS_URI = "http://www.w3.org/XML/1998/namespace"
STREAM_NS_URI = "http://etherx.jabber.org/streams"
STREAM_ERROR_NS_URI = "urn:ietf:params:xml:ns:xmpp-streams"
STANZA_ERROR_NS_URI = "urn:ietf:params:xml:ns:xmpp-stanzas"
CLIENT_NS_URI = "jabber:client"
SERVER_NS_URI = "jabber:server"
